# -*- coding: utf-8 -*-
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
#cython: embedsignature=True


import numpy as np
cimport numpy as np
oldsettings = np.seterr(divide='ignore')

from .numpy_types import typedict
from .numpy_types cimport np_floating, np_numeric
from .array_descriptors import ArrayDescriptor

from .configure_logger import get_logger

cimport cython
cdef extern from "numpy/npy_math.h":
    bint npy_isnan(double x) nogil


cdef class ArrayShapes(object):
    """
    Base Class for a Cython cdef class which helps to handle
    memoryviews better
    """
    def __cinit__(self, *args, **kwargs):
        """init the file"""
        if not hasattr(self, '__module__'):
            msg = "don't instantiate cdef class directly, please subclass in python class"
            raise NotImplementedError(msg)
        self.dtypes = {}
        for cls in self.__class__.__mro__:
            self._search_memview(cls)

    def __init__(self, *args, **kwargs):
        """
        inits the Array and creates the constands for NAN and NINF
        """
        # super class has to be called even if the super class of ArrayShapes
        # is only `object`
        super().__init__(*args, **kwargs)
        # set NAN-Values
        self.NAN_f = np.NAN #np.float32(0) / np.float32(0)
        self.INF_f = np.float32(1) / np.float32(0)
        self.NINF_f = np.float32(-1) / np.float32(0)

        self.NAN_d = np.NAN #np.float64(0) / np.float64(0)
        self.INF_d = np.float64(1) / np.float64(0)
        self.NINF_d = np.float64(-1) / np.float64(0)

        # create Class logger
        self.logger = get_logger(self)

    cdef char _isnan(self, np_floating x) nogil:
        """
        check for nan
        """
        return npy_isnan(x)

    def isnan_py(self, np_numeric x) -> bool:
        """python wrapper around _isnan()

        Parameters
        ----------
        x:
            the value to test for is_nan"""
        return bool(self._isnan(float(x)))

    @cython.initializedcheck(False)
    cpdef _search_memview(self, cls):
        """
        Search the memoryviews in the class cls
        and determine their dtype by trial and error

        Parameters
        ----------
        cls : the subclass to check for memorytypes
        """
        for name, attr in cls.__dict__.items():
            if not name.startswith('__'):
                if name.startswith('_'):
                    try:
                        attr = getattr(self, name)
                        ismemview = [cl.__name__.endswith('memoryview')
                                     for cl in attr.__class__.mro()]
                    except AttributeError as err:
                        if str(err) == 'Memoryview is not initialized':
                            ismemview = [True]
                        else:
                            raise
                    if any(ismemview):
                        prop_name = name[1:]
                        self._init_array(prop_name, (0,))

    @cython.initializedcheck(False)
    def _init_array(self, name, shape0=None, default=None):
        """
        Inits the attribute name with an empty array with the specified shape
        the dtype is inferred from the definition of the attribute

        Parameters
        ----------
        name : str
            the name of the array (without leading underscore)
        shape0 : tuple, optional
            the shape of the array that should be initialized
        default : number, optional
            the default value to fill the created array
            if None, an empty array is created
        """
        intern_name = '_%s' % name
        if shape0 is None:
            shape = (0,)
        elif not isinstance(shape0, tuple):
            shape = (int(shape0), )
        else:
            shape = shape0
        ndim = len(shape)
        np_dtype = 'f4'
        arr = np.empty(shape, dtype=np_dtype)
        try:
            setattr(self, intern_name, arr)

        except ValueError as err:
            msg = str(err)
            if msg.startswith('Buffer has wrong number of dimensions'):
                arr = self._change_ndims(msg, shape, np_dtype)
                ndim = arr.ndim
                try:
                    setattr(self, intern_name, arr)
                except ValueError as err:
                    msg = str(err)
                    if msg.startswith('Buffer dtype mismatch'):
                        arr, np_dtype = self._change_dtype(msg, arr.shape,
                                                           default)
                        ndim = arr.ndim
                        setattr(self, intern_name, arr)
                    else:
                        raise

            elif msg.startswith('Buffer dtype mismatch'):
                arr, np_dtype = self._change_dtype(msg, arr.shape, default)
                #print intern_name, arr.dtype, np_dtype, arr.itemsize
                #print msg
                try:
                    setattr(self, intern_name, arr)
                except ValueError as err:
                    print(err)
                    raise err
            else:
                raise
        self.dtypes[name] = ArrayDescriptor(name, np_dtype, ndim)

    def _change_ndims(self, msg, shape, np_dtype):
        """
        according to the error message received the number of dimensions
        is changed and a new array returned
        Parameters
        ----------
        msg : str
            the message raised by the Exeption
        shape : tuple
            the shape of the array to create
        np_dtype : str
            the numpy-dtype of the array to create

        Returns
        -------
        arr : np.array
            the data with the changed datatype
        """
        cndim = int(msg.split("expected ")[1].split(", got")[0])
        shape = tuple([0] * cndim)
        arr = np.empty(shape, dtype=np_dtype)
        return arr

    def _change_dtype(self, msg: str, shape, default):
        """
        according to the error message received the dtype is changed
        and a new array returned

        Parameters
        ----------
        msg : str
            the message raised by the Exeption
        shape : tuple
            the shape of the array to create
        default : number
            the default value to fill the array

        Returns
        -------
        arr : np.array
            the data with the changed datatype
        np_dtype : str
            the numpy datatype
        """
        cdtype = msg.split("expected '")[1].split("' but got")[0]
        np_dtype = typedict[cdtype]
        arr = np.empty(shape, dtype=np_dtype)
        if default is not None:
            arr.fill(default)
        return arr, np_dtype



# -*- coding: utf-8 -*-

from cythonarrays.numpy_types cimport *

from cythonarrays.array_shapes cimport ArrayShapes
from cythonarrays.array_shapes import ArrayShapes


cdef class _Example(ArrayShapes):
    """
    Cython CDefClass for Example model
    with coordinates groups and zones
    """
    cdef public long32 groups
    cdef public long32 origins
    cdef public long32 destinations
    cdef public char n_threads

    cdef public ARRAY_1D_i4 _zonenumbers_i

    # parameter of groups
    cdef public ARRAY_1D_d _param_g

    # distance matrix (km from a to b)
    cdef public ARRAY_2D_d _km_ij

    # persons per group and zone
    cdef public ARRAY_2D_d _persons_gi
    # jobs per zone
    cdef public ARRAY_1D_d _jobs_j

    # resulting trip matrix
    cdef public ARRAY_2D_d _trips_ij

    # not initialized matrix
    cdef public ARRAY_2D_i4 _not_initialized_ij

    # boolen array
    cdef public ARRAY_1D_i1 _valid_g
    cdef public ARRAY_1D_u1 _invalid_g

    cpdef char calc_model(self) except -1
    cdef double _calc_weight_destination(self, double param,
                                         double minutes, double jobs) nogil
    cdef ARRAY_1D_d _calc_p_destination(self, long32 g) nogil





# -*- coding: utf-8 -*-
cimport numpy as np
cimport cython

from .numpy_types cimport np_floating

cdef class ArrayShapes(object):
    cdef public float NAN_f
    cdef public float INF_f
    cdef public float NINF_f

    cdef public double NAN_d
    cdef public double INF_d
    cdef public double NINF_d

    cdef char _isnan(self, np_floating x) nogil

    cpdef _search_memview(self, cls)


# -*- coding: utf-8 -*-

from typing import Tuple, Union
import numpy as np
import xarray as xr
from cythonarrays.array_descriptors import ArrayDescriptor
from cythonarrays.configure_logger import get_logger


class _ArrayProperties:
    """
    Base Class for Cython CDef Class

    :meta public:
    """
    _coordinates = {}

    def __init__(self, *args, **kwargs):
        """
        create properties for all arrays
        """
        for descr in self.dtypes.values():
            prop = self._create_prop(descr)
            setattr(self.__class__, descr.name, prop)
            # create Class logger
            self.logger = get_logger(self)

    def _create_prop(self, descr: ArrayDescriptor):
        """
        Create the property name that reads and writes
        the internal attribute _name

        Parameters
        ----------
        descr:
            the Array Description
        """
        intern_name = '_%s' % descr.name

        def fget(self):
            arr = getattr(self, intern_name)
            return np.array(arr, copy=False).view(dtype=descr.dtype_numpy)

        def fset(self, value):
            self.set_array(descr.name, value)

        def fdel(self):
            self.set_array(descr.name, np.empty(tuple([0]*descr.ndim)))

        fdoc = str(descr)
        prop = property(fget, fset, fdel, fdoc)
        return prop

    def init_array(self,
                   name: str,
                   shape: Tuple[Union[str, int]]=None,
                   default: Union[int, float]=None):
        """
        Inits the attribute name with an empty array with the specified shape
        and fill with the default value

        Parameters
        ----------
        name:
            the name of the attribute
        shape:
            the shape of the dtype. If not given,
            the dtype stored in self.dtypes is used.
            if None, an array of shape [0] * ndim is initialized
        default:
            the default value. If not given,
            the default value stored in self.dtypes is used.
            if None, an empty array is initialized
        """
        descr = self.dtypes[name]
        if shape is not None:
            descr.shape = shape
        if default is not None:
            descr.default = default
        if descr.shape is not None:
            self.check_ndims(descr)
            target_shape = descr.get_shape(self)
            arr = np.empty(target_shape, dtype=descr.dtype)
            if descr.default is not None:
                arr.fill(descr.default)
        else:
            arr = np.empty([0] * descr.ndim, dtype=descr.dtype)
        self.set_array(name, arr)

    def set_array(self,
                  name: str,
                  value: Union[int, float, list, np.ndarray],
                  shape: Tuple[Union[int, str]]=None):
        """
        Sets the attribute name to the value and casts to the correct dtype
        if necessary

        Parameters
        ----------
        name:
            the name of the attribute to set
        value: Array-like
            the value to assign to the memoryview
        shape:
            a tuple of ints or str to validate the shape of the array provided

        Examples
        --------
          >>> arr = np.array([5, 7, 5, 8], dtype='u1')
          >>> shape = ('n_agents', )
          >>> model.n_agents
          4
          >>> model.set_array('age', arr, shape)
          >>> model.n_agents = 5
          >>> model.set_array('age', arr, shape)
          Exception...
          >>> shape = (4, )
          >>> model.set_array('age', arr, shape)
        """
        intern_name = '_%s' % name
        descr = self.dtypes[name]
        if shape is not None:
            descr.shape = shape
            self.check_ndims(descr)
        arr = descr.validate_array(value, self)
        setattr(self, intern_name, arr)

    def reset_array(self, name: str):
        """
        Reset array to its default value

        Parameters
        ----------
        name:
            the name of the array

        """
        descr = self.dtypes[name]
        default = descr.default
        getattr(self, name).fill(default)

    def check_ndims(self, descr: ArrayDescriptor):
        """
        Check the number of dimensions

        Parameters
        ----------
        descr :
            the Array Descriptor
        """
        if len(descr.shape) != descr.ndim:
            msg = '{ndim} Dimensions required, shape {s} has {n} dimensions'
            raise ValueError(msg.format(ndim=descr.ndim,
                                        s=descr.shape, n=len(descr.shape)))

    def init_arrays(self):
        """Initialize all arrays defined in self.dtypes"""
        for name in self.dtypes:
            self.init_array(name)

    def init_object_array(self, name: str, shape: Union[str, Tuple[str, int]]):
        """
        initialize an non-cython array to assure the consistency of data

        Parameters
        ----------
        name : str
            the name of the attribute
        shape : tuple or str
            the shape of the array
        """
        if not isinstance(shape, (tuple, list)):
            shape = shape.split(',')
        ndim = len(shape)

        intern_name = '_%s' % name
        self.dtypes[name] = descr = ArrayDescriptor(name, 'O', ndim)
        prop = self._create_prop(descr)
        setattr(self.__class__, descr.name, prop)
        self.init_array(name, shape=shape)

    def create_ds(self):
        """
        Create xarray-Dataset that is linked to the arrays
        """
        self.ds = xr.Dataset()
        for name, dtype in self.dtypes.items():
            arr = getattr(self, name)
            self.ds[name] = xr.DataArray(arr, dims=dtype._shape)
        coord_dict = {}
        for key, value in self._coordinates.items():
            coord_dict[key] = getattr(self, value)
        self.ds = self.ds.assign_coords(**coord_dict)

    def save_dataset_to_netcdf(self, filepath: str):
        """
        Save Dataset to netcdf-file

        Parameters
        ----------
        filepath:
            the filepath to store the data
        """
        if not hasattr(self, 'ds'):
            self.create_ds()
        self.ds.to_netcdf(filepath)

    @classmethod
    def from_netcdf(cls, filepath: str) -> '_ArrayProperties':
        """
        Read Data from a netcdf-file and create a new Cdef-Class-instance

        This works only, if the dimensions of the cythonarrays-class
        are specified in the __init__() of the subclass
        and the argument names match the names of the dimensions

        Parameters
        ----------
        filepath:
            the filepath to read the netcdf-data from
        """
        ds = xr.open_dataset(filepath)
        # create a dictionary with the dimensions
        dimensions = dict()
        cls_init_args = cls.__init__.__code__.co_varnames
        for key, value in cls._coordinates.items():
            # only if these dimensions are arguments to the __init__-method
            if key in cls_init_args:
                dimensions[key] = len(ds[value])
        # create the class instance
        self = cls(**dimensions)
        # link the Dataset and the Arrays
        self.ds = ds
        for name, dtype in self.dtypes.items():
            arr = ds[name]
            setattr(self, name, arr.values)
        return self



# -*- coding: utf-8 -*-

import multiprocessing
import xarray as xr
import os
from collections import defaultdict
import numpy as np

from cythonarrays.array_properties import _ArrayProperties
import pyximport; pyximport.install()
from .example_cython import (_Example, DestinationChoiceError)

np.seterr(divide='ignore', invalid='ignore')


class Example(_Example, _ArrayProperties):
    """
    Example job choice model

    Parameters
    ----------
    groups : int
        the number of groups
    origins : int
        the number of zones
    threading: bool, optional(default=True)
        use multiprocessing?
    init_arrays: bool, optional(default=True)
        if true, the arrays are initialized in the beginning

    """
    _coordinates = {'origins': 'zonenumbers_i',
                    'destinations': 'zonenumbers_i',
                    'groups': 'groupnames_g',
                    }

    def __init__(self,
                 groups,
                 origins,
                 threading=True,
                 init_arrays=True):
        super().__init__()

        self.groups = groups
        zones = origins
        self.origins = zones
        self.destinations = zones
        self.set_n_threads(threading)

        self.define_arrays()
        if init_arrays:
            self.init_arrays()

    def set_n_threads(self, threading=True):
        """Set the number of threads"""
        if threading:
            self.n_threads = min(multiprocessing.cpu_count(), self.groups)
        else:
            self.n_threads = 1

    def define_arrays(self):
        """Define the arrays"""
        self.init_array('param_g', 'groups', -0.1)
        self.init_array('km_ij', 'origins, destinations')

        self.init_array('persons_gi', 'groups, origins')
        self.init_array('jobs_j', 'destinations')
        self.init_array('trips_ij', ['origins', 'destinations'], 0)

        self.init_array('zonenumbers_i', 'origins')
        self.init_object_array('groupnames_g', 'groups')
        self.init_object_array('zonenames_i', ['origins'])

        self.init_array('valid_g', 'groups', 1)
        self.init_array('invalid_g', 'groups', 0)

        # the following array will not be initialized for testing purposes
        # self.init_array('not_initialized_ij', 'zones, groups')

